/*
 * JPEG huffman parsing and repairing for the pac7311 usb webcam chip.
 *
 * Copyright (c) 2006, Jörg Schummer <joerg.schummer@web.de>
 *
 *
 * Large parts of this code are taken from the
 *
 * Small jpeg decoder library
 *
 * Copyright (c) 2006, Luc Saillard <luc@saillard.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *
 * - Neither the name of the author nor the names of its contributors may be
 *  used to endorse or promote products derived from this software without
 *  specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>

#include "pac7311_jpeg.h"
#include "pac7311_huffman.h"
#include "pac7311_jpeg_header.h"

#define cY	1
#define cCb	2
#define cCr	3


#define SEG_TYPE_START 99
#define SEG_LEN_END      -1
#define SEG_LEN_VARIABLE -1

struct {

	int type;
	unsigned long len;
	
} pac7311_seg_len[] = {

	{ SEG_TYPE_START,	1018 },
	{ 01,			1024 },
	{ 02,			512 },
	{ 00,			SEG_LEN_VARIABLE },
	{ SEG_LEN_END }
};

//static struct jdec_private *priv;

int pac7311_jpeg_check_segment_len(unsigned char seg_type, unsigned long len)
{
	int i = 0;
	
	while ((pac7311_seg_len[i].type != SEG_LEN_END) &&
	       (pac7311_seg_len[i].type != seg_type))
	{
		++i;
	}
	
	if (pac7311_seg_len[i].type == seg_type)
	{
		return (((len == pac7311_seg_len[i].len)) ||
		        (pac7311_seg_len[i].len == SEG_LEN_VARIABLE));
	}
	else
	{
		// return false for unknown segments
		return 0;
	}
}




/*
 * 4 functions to manage the stream
 *
 *  fill_nbits: put at least nbits in the reservoir of bits.
 *              But convert any 0xff,0x00 into 0xff
 *  get_nbits: read nbits from the stream, and put it in result,
 *             bits is removed from the stream and the reservoir is filled
 *             automaticaly. The result is signed according to the number of
 *             bits.
 *  look_nbits: read nbits from the stream without marking as read.
 *  skip_nbits: read nbits from the stream but do not return the result.
 * 
 * stream: current pointer in the jpeg data (read bytes per bytes)
 * nbits_in_reservoir: number of bits filled into the reservoir
 * reservoir: register that contains bits information. Only nbits_in_reservoir
 *            is valid.
 *                          nbits_in_reservoir
 *                        <--    17 bits    -->
 *            Ex: 0000 0000 1010 0000 1111 0000   <== reservoir
 *                        ^
 *                        bit 1
 *            To get two bits from this example
 *                 result = (reservoir >> 15) & 3
 *
 */
 
 
static inline void fill_nbits(struct jdec_private *priv, unsigned char nbits_wanted)
{
	unsigned char c, cnext;

	if (priv->eof)
		return;

	while (priv->nbits_in_reservoir < nbits_wanted)
	{
		if (priv->stream >= priv->stream_end)
		{
			priv->eof = 1;
			return;
		}

		c = *(priv->stream++);

		if (priv->stream < priv->stream_end)
		{
			/* AAXX
			     ^  */
			     
			cnext = *priv->stream;

			if ((c == 0xff) &&
			    (priv->stream < (priv->stream_end - 3)) &&
			    (!memcmp(priv->stream, "\xff\xff", 2)))
			{
				const unsigned char *seg_end = (priv->stream - 1);
			
				const unsigned char *next_seg_start = priv->stream_end;
				unsigned char next_seg_type = 0xff;
			
				/* AAFFFFXX
				     ^      */

				priv->stream += 2;

				/* AAFFFFXX
				         ^   */

				if ((*priv->stream == 0xff) &&
				    (priv->stream < (priv->stream_end - 3)))
				{
					++seg_end;
				
					/* AAFFFFFFXXXX
					         ^      */
				
					next_seg_type = *(++priv->stream);
				
					/* AAFFFFFFXXXX
					           ^    */
					
					next_seg_start = ++priv->stream;

					cnext = *priv->stream;
				
					/* AAFFFFFFXXBB
					             ^  */
				}
				else
				{
					/* AAFFFFXX
					         ^   */

					if (priv->stream < (priv->stream_end - 3))
					{
						/* FFFFFFXXXXXX
						         ^      */
						
						next_seg_type = *(priv->stream++);  
						       
						/* FFFFFFXXXXXX
						           ^    */
						
						next_seg_start = priv->stream;
						
						c = *(priv->stream++);
						
						cnext = *priv->stream;

						/* FFFFFFXXAABB
						             ^  */
					}
				}									

				if (!pac7311_jpeg_check_segment_len(
					priv->cur_seg_type,
					seg_end - priv->cur_seg_start))
				{
					priv->stream = next_seg_start - 4;
				
					longjmp(priv->jmp_state, -EIO);
				}
				else
				{
					priv->cur_seg_type = next_seg_type;
					priv->cur_seg_start = next_seg_start;
				}
			}
		}
		else
			cnext = 0xff;

		if (c == 0xff)
		{
			if (cnext == 0x00)

				++priv->stream;
			/*

			else if (cnext != 0xff)
				fprintf(stderr, "tinyjpeg: ERROR: unknown JPEG marker FF%02X at %d\n",
				        cnext, (priv->stream - priv->stream_begin));
            */
		}

		priv->reservoir <<= 8;
		priv->reservoir |= c;
		priv->nbits_in_reservoir += 8;
	}
}



static inline short int get_nbits(struct jdec_private *priv, unsigned char nbits_wanted)
{
	short int result;
	
	fill_nbits(priv, nbits_wanted);
	result = (priv->reservoir >> (priv->nbits_in_reservoir - nbits_wanted));

	priv->nbits_in_reservoir -= nbits_wanted;
	priv->reservoir &= ((1U << priv->nbits_in_reservoir) - 1);

	if ((unsigned int) result < (1UL << (nbits_wanted - 1)))
		result += (0xFFFFFFFFUL << nbits_wanted) + 1;
		
	return result;
}


static inline unsigned int look_nbits(struct jdec_private *priv, unsigned char nbits_wanted)
{
	fill_nbits(priv, nbits_wanted);
	
	return (priv->reservoir >> (priv->nbits_in_reservoir - nbits_wanted));
}


/* To speed up the decoding, we assume that the reservoir have enough bit 
*/
static inline void skip_nbits(struct jdec_private *priv, unsigned char nbits_wanted)
{
	priv->nbits_in_reservoir -= nbits_wanted;
	priv->reservoir &= ((1U << priv->nbits_in_reservoir) - 1);
}




inline void write_bits_flush_output_byte(struct jdec_private *priv, unsigned char c)
{
	*(priv->out_stream++) = c;
	
	if (c == 0xff)
	{
		*(priv->out_stream++) = 0x00;
	}
}


inline void write_bits_flush(struct jdec_private *priv)
{
	int i;

	for (i = 0; i < BITBUF_BIT_SIZE; i += 8)
	{
		if (priv->nbits_in_out_reservoir > i)
		{
			write_bits_flush_output_byte(priv,
				(unsigned char) (priv->out_reservoir >> (BITBUF_BIT_SIZE - i - 8)));
		}
		else
		{
			break;
		}
	}

	priv->out_reservoir = 0;
	priv->nbits_in_out_reservoir = 0;
}


inline void write_bits(struct jdec_private *priv, BITBUF_TYPE value, unsigned char bits)
{
	char shift = BITBUF_BIT_SIZE - priv->nbits_in_out_reservoir - bits;
	
	value &= (((BITBUF_TYPE) 1) << bits) - 1;
	
	if (shift > 0)
	{
		priv->out_reservoir |= (value << shift);
		priv->nbits_in_out_reservoir += bits;
	}
	else
	{
		shift *= -1;
		priv->out_reservoir |= (value >> shift);
		priv->nbits_in_out_reservoir = BITBUF_BIT_SIZE;

		write_bits_flush(priv);
		
		if (shift > 0)
		{
			priv->out_reservoir = (value << (BITBUF_BIT_SIZE - shift));
			priv->nbits_in_out_reservoir = shift;
		}
	}
}


static inline void write_bits_copy_skip(struct jdec_private *priv, unsigned char len)
{
	unsigned int buf;
	
	buf = look_nbits(priv, len);
	
	if (!priv->eof)
	{
		write_bits(priv, buf, len);
		skip_nbits(priv, len);
	}
}



static int get_copy_next_huffman_code(struct jdec_private *priv, struct huffman_table *huffman_table)
{
	int value, hcode;
	unsigned int extra_nbits, nbits;
	uint16_t *slowtable;

	hcode = look_nbits(priv, HUFFMAN_HASH_NBITS);
	if (priv->eof)
		return -2;

	value = huffman_table->lookup[hcode];
	if (value >= 0)
	{
		unsigned int code_size = huffman_table->code_size[value];
		write_bits_copy_skip(priv, code_size);

		return value;
	}

	/* Decode more bits each time ... */
	for (extra_nbits = 0; extra_nbits < (16 - HUFFMAN_HASH_NBITS); ++extra_nbits)
	{
		nbits = HUFFMAN_HASH_NBITS + 1 + extra_nbits;

		hcode = look_nbits(priv, nbits);
		if (priv->eof)
			return -2;

		slowtable = huffman_table->slowtable[extra_nbits];

		/* Search if the code is in this array */
		while (slowtable[0])
		{
			if (slowtable[0] == hcode)
			{
				write_bits_copy_skip(priv, nbits);

				return slowtable[1];
			}

			slowtable += 2;
		}
	}

	fprintf(stderr, "tinyjpeg: ERROR: Unknown Huffman-Code!\n");

	return -1;
}

int get_copy_Huffman_data_unit(struct jdec_private *priv, int component)
{
	unsigned char j;
	int huff_code;
	int bits_parsed;
	unsigned char size_val, count_0;
	
	struct component *c = &priv->component_infos[component];

	/* DC coefficient decoding */
	huff_code = get_copy_next_huffman_code(priv, c->DC_table);
    bits_parsed = huff_code;

	if (huff_code < 0)
	{
		write_bits(priv, c->DC_table->zero_code, c->DC_table->zero_code_size);
	}
	else if (huff_code)
	{
		write_bits_copy_skip(priv, huff_code);
		if (priv->eof)
		{
			write_bits(priv, 0, huff_code);
		}
	}

	/* AC coefficient decoding */
	j = 1;
	while (j < 64)
	{
		huff_code = get_copy_next_huffman_code(priv, c->AC_table);
        bits_parsed += huff_code;
		if (huff_code < 0)
		{
			write_bits(priv, c->AC_table->zero_code, c->AC_table->zero_code_size);
			huff_code = 0;
		}

		size_val = huff_code & 0xF;
		count_0 = huff_code >> 4;

		if (size_val == 0)
		{ /* RLE */
			if (count_0 == 0)
			{
				break;	/* EOB found, go out */
			}
			else if (count_0 == 0xF)
			{
				j += 16;	/* skip 16 zeros */
			}
		}
		else
		{
			j += count_0;	/* skip count_0 zeroes */

			write_bits_copy_skip(priv, size_val);
			if (priv->eof)
			{
				write_bits(priv, 0, size_val);
			}

			j++;
		}
	}
    return bits_parsed;
}

unsigned char *pac7311_jpeg_get_header(int image_width, int image_height)
{
	pac7311_jpeg_header[PAC7311_JPEG_HEADER_SIZE_POS + 0] = (unsigned char) (image_height >> 8);
	pac7311_jpeg_header[PAC7311_JPEG_HEADER_SIZE_POS + 1] = (unsigned char) (image_height & 0xff);
	pac7311_jpeg_header[PAC7311_JPEG_HEADER_SIZE_POS + 2] = (unsigned char) (image_width >> 8);
	pac7311_jpeg_header[PAC7311_JPEG_HEADER_SIZE_POS + 3] = (unsigned char) (image_width & 0xff);

	return pac7311_jpeg_header;
}

int pac7311_jpeg_decode(unsigned int width, unsigned int height,
                          unsigned char *jpeg_stream, unsigned int jpeg_stream_size)
{
	unsigned char c;
	unsigned long count, num_blocks;
    int bits_parsed;
    int countmcu = 0;
	struct jdec_private *priv;

	priv = (struct jdec_private *) calloc(1, sizeof(struct jdec_private));

	priv->reservoir = 0;
	priv->nbits_in_reservoir = 0;
	
	priv->out_reservoir = 0;
	priv->nbits_in_out_reservoir = 0;

	priv->eof = 0;
	
	priv->cur_seg_type = SEG_TYPE_START;
	
	priv->out_stream = priv->out_stream_begin = (unsigned char *) malloc(jpeg_stream_size);

    priv->mcumarker = *jpeg_stream;
	priv->width = width;
	priv->height = height;
	
	priv->stream = priv->stream_begin = jpeg_stream;
	priv->stream_end = jpeg_stream + jpeg_stream_size;
	priv->stream_length = jpeg_stream_size;
	
	priv->cur_seg_start = priv->stream;
	
	priv->HTDC[0] = huffman_table_pac7311_DC0;
	priv->HTDC[1] = huffman_table_pac7311_DC1;
	priv->HTAC[0] = huffman_table_pac7311_AC0;
	priv->HTAC[1] = huffman_table_pac7311_AC1;
	
	priv->component_infos[cY].DC_table =  &huffman_table_pac7311_DC0;
	priv->component_infos[cY].AC_table =  &huffman_table_pac7311_AC0;
	priv->component_infos[cCb].DC_table = &huffman_table_pac7311_DC1;
	priv->component_infos[cCb].AC_table = &huffman_table_pac7311_AC1;
	priv->component_infos[cCr].DC_table = &huffman_table_pac7311_DC1;
	priv->component_infos[cCr].AC_table = &huffman_table_pac7311_AC1;

	if (((priv->width % 16) != 0) || ((priv->height % 8)))
	{
		fprintf(stderr, "tinyjpeg_write_parse_marker: width must be divisible by 16, height by 8\n");
		return 1;
	}

	if (setjmp(priv->jmp_state))
	{
		fprintf(stderr, "tinyjpeg_write_parse_marker: segment marker fault, segment 0x%02x had length %ld\n", priv->cur_seg_type, (unsigned long) (priv->stream - priv->cur_seg_start));
		return 2;
	}

	num_blocks = (priv->width / 16) * (priv->height / 8);

	for (count = 0; count < num_blocks; ++count)
	{
		c = look_nbits(priv, 8);

		if ((c != priv->mcumarker) && (!priv->eof)) {
			fprintf(stderr, "tinyjpeg_write_parse_marker: mcu marker 0x%02x missing 8 Bit = 0x%X | %i MCUs parsed\n", priv->mcumarker, c, countmcu);
        	//return 2;
		}

		skip_nbits(priv, 8);

		bits_parsed = get_copy_Huffman_data_unit(priv, cY);
		bits_parsed += get_copy_Huffman_data_unit(priv, cY);
		bits_parsed += get_copy_Huffman_data_unit(priv, cCb);
		bits_parsed += get_copy_Huffman_data_unit(priv, cCr);

        countmcu++;
	}

	if (!pac7311_jpeg_check_segment_len(priv->cur_seg_type, (unsigned long) (priv->stream - priv->cur_seg_start)))
	{
		longjmp(priv->jmp_state, -EIO);
	}

	write_bits_flush(priv);

    memcpy(jpeg_stream, pac7311_jpeg_header, PAC7311_JPEG_HEADER_SIZE);
    memcpy((jpeg_stream + PAC7311_JPEG_HEADER_SIZE), priv->out_stream_begin, (priv->out_stream - priv->out_stream_begin));
	free(priv->out_stream_begin);
	free(priv);
	return 0;
}
